﻿using System.Diagnostics;
using System.IO;
using SharpDX;
using SharpDX.Direct2D1;
using SharpDX.Direct2D1.Effects;
using SharpDX.IO;
using D3D = SharpDX.Direct3D11;
using DW = SharpDX.DirectWrite;
using DXGI = SharpDX.DXGI;
using WIC = SharpDX.WIC;

namespace Cyjb.Demos.Direct2DBevel
{
	/// <summary>
	/// C# 使用 Direct2D 实现斜角效果 Demo。
	/// 文章链接：http://www.cnblogs.com/cyjb/p/Direct2DBevelEffect.html
	/// by CYJB http://www.cnblogs.com/cyjb/。
	/// </summary>
	class Program
	{
		static void Main()
		{
			// ===== 初始化所有需要的设备和工厂 =====
			// 初始化 D3D 设备。
			var defaultDevice = new D3D.Device(SharpDX.Direct3D.DriverType.Hardware, D3D.DeviceCreationFlags.BgraSupport);
			// 获取 Direct3D 11.1 设备。
			var d3dDevice = defaultDevice.QueryInterface<D3D.Device1>();
			// 获取 DXGI 设备。
			var dxgiDevice = d3dDevice.QueryInterface<DXGI.Device>();
			// 初始化 Direct2D 设备。
			var d2dDevice = new Device(dxgiDevice);
			// 初始化 WIC 工厂。
			var imagingFactory = new WIC.ImagingFactory2();
			// 初始化 Direct2D 设备上下文。
			var d2dContext = new DeviceContext(d2dDevice, DeviceContextOptions.None);
			// 初始化 DirectWrite 工厂。
			var dwFactory = new DW.Factory(DW.FactoryType.Shared);

			// ===== 参数设置 =====
			// 位图的像素格式。
			var d2PixelFormat = new PixelFormat(DXGI.Format.R8G8B8A8_UNorm, AlphaMode.Premultiplied);
			// 位图属性。
			var d2dBitmapProps = new BitmapProperties1(d2PixelFormat, 96, 96, BitmapOptions.Target);
			// WIC 像素格式。
			var wicPixelFormat = WIC.PixelFormat.Format32bppPRGBA;

			// 源位图的尺寸。
			Size2F sourceBitmapSize = new Size2F(250.0f, 150.0f);
			// 位图中大矩形的尺寸。
			Size2F largeRectangleSize = new Size2F(200.0f, 100.0f);
			// 位图中小矩形的尺寸。
			Size2F smallRectangleSize = new Size2F(150.0f, 70.0f);
			// 圆角的半径。
			float roundingRadius = 15.0f;
			// 位图中大矩形的线条宽度。
			float strokeWidth = 10.0f;
			// 字体的尺寸。
			float fontSize = 32.0f;

			// 高斯模糊特效的参数。
			float gaussianBlurStDev = 8.0f;
			// 点光照特效的参数。
			Vector3 specularLightPosition = new Vector3(300.0f, -1000.0f, 3000.0f);
			float specularExponent = 20.0f;
			float specularConstant = 0.75f;
			float specularSurfaceScale = 5.0f;
			Vector4 arithmeticCoefficients = new Vector4(0.0f, 1.0f, 1.0f, 0.0f);

			// 创建 Direct2D 资源。
			var whiteBrush = new SolidColorBrush(d2dContext, Color.White);
			var redBrush = new SolidColorBrush(d2dContext, Color.Red);
			var textFormat = new DW.TextFormat(dwFactory, "Segoe UI", fontSize);
			textFormat.TextAlignment = DW.TextAlignment.Center;
			textFormat.ParagraphAlignment = DW.ParagraphAlignment.Center;

			// ===== 绘制图形的源位图 =====
			var bitmap = new Bitmap1(d2dContext, new Size2((int)sourceBitmapSize.Width, (int)sourceBitmapSize.Height),
				d2dBitmapProps);
			// 在源位图上绘制图形。
			d2dContext.Target = bitmap;
			d2dContext.BeginDraw();
			d2dContext.Clear(Color.Transparent);

			// 绘制大矩形。
			RoundedRectangle largeRoundedRect = new RoundedRectangle()
			{
				Rect = new RectangleF(
					(sourceBitmapSize.Width - largeRectangleSize.Width) / 2,
					(sourceBitmapSize.Height - largeRectangleSize.Height) / 2,
					largeRectangleSize.Width,
					largeRectangleSize.Height),
				RadiusX = roundingRadius,
				RadiusY = roundingRadius
			};
			d2dContext.DrawRoundedRectangle(largeRoundedRect, redBrush, strokeWidth);

			// 绘制小矩形。
			RectangleF smallRect = new RectangleF(
				(sourceBitmapSize.Width - smallRectangleSize.Width) / 2,
				(sourceBitmapSize.Height - smallRectangleSize.Height) / 2,
				smallRectangleSize.Width,
				smallRectangleSize.Height);
			RoundedRectangle smallRoundedRect = new RoundedRectangle()
			{
				Rect = smallRect,
				RadiusX = roundingRadius,
				RadiusY = roundingRadius
			};
			d2dContext.FillRoundedRectangle(smallRoundedRect, redBrush);

			// 绘制文本。
			string m_string = "Direct2D";
			d2dContext.DrawText(m_string, textFormat, smallRect, whiteBrush);

			d2dContext.EndDraw();

			// ===== 配置特效 =====
			// 高斯模糊特效。
			var gaussBlurEffect = new GaussianBlur(d2dContext);
			gaussBlurEffect.SetInput(0, bitmap, true);
			gaussBlurEffect.StandardDeviation = gaussianBlurStDev;
			// 点光照特效。
			var pointSpecularEffect = new PointSpecular(d2dContext);
			pointSpecularEffect.SetInputEffect(0, gaussBlurEffect);
			pointSpecularEffect.LightPosition = specularLightPosition;
			pointSpecularEffect.SpecularExponent = specularExponent;
			pointSpecularEffect.SpecularConstant = specularConstant;
			pointSpecularEffect.SurfaceScale = specularSurfaceScale;
			// 组合特效。
			var composite = new Composite(d2dContext);
			composite.SetInput(0, bitmap, true);
			composite.SetInputEffect(1, pointSpecularEffect);
			composite.Mode = CompositeMode.SourceIn;
			// 算数组合特效。
			var aritheticComposite = new ArithmeticComposite(d2dContext);
			aritheticComposite.SetInput(0, bitmap, true);
			aritheticComposite.SetInputEffect(1, composite);
			aritheticComposite.Coefficients = arithmeticCoefficients;

			// ===== 绘制目标图形 =====
			var d2dRenderTarget = new Bitmap1(d2dContext,
				new Size2((int)(sourceBitmapSize.Width), (int)(sourceBitmapSize.Height * 3)),
				d2dBitmapProps);
			// 在目标位图上绘制图形。
			d2dContext.Target = d2dRenderTarget;
			d2dContext.BeginDraw();
			d2dContext.Clear(Color.Transparent);

			d2dContext.DrawImage(bitmap, new Vector2(0, 0));
			d2dContext.DrawImage(composite, new Vector2(0, sourceBitmapSize.Height));
			d2dContext.DrawImage(aritheticComposite, new Vector2(0, sourceBitmapSize.Height * 2));

			d2dContext.EndDraw();

			// ====== 保存绘制结果 =====
			// 删除已存在的输出文件。
			string outputPath = "Output.png";
			if (File.Exists(outputPath))
			{
				File.Delete(outputPath);
			}
			// 将位图编码并输出。
			using (var stream = new WIC.WICStream(imagingFactory, outputPath, NativeFileAccess.Write))
			{
				// 选择编码为 PNG 文件。
				using (var encoder = new WIC.PngBitmapEncoder(imagingFactory))
				{
					encoder.Initialize(stream);
					// 帧编码器。
					using (var bitmapFrameEncode = new WIC.BitmapFrameEncode(encoder))
					{
						bitmapFrameEncode.Initialize();
						bitmapFrameEncode.SetSize(d2dRenderTarget.PixelSize.Width, d2dRenderTarget.PixelSize.Height);
						bitmapFrameEncode.SetPixelFormat(ref wicPixelFormat);
						// 将位图编码。
						using (var imageEncoder = new WIC.ImageEncoder(imagingFactory, d2dDevice))
						{
							imageEncoder.WriteFrame(
								d2dRenderTarget,
								bitmapFrameEncode,
								new WIC.ImageParameters(d2PixelFormat, 96, 96, 0, 0,
									d2dRenderTarget.PixelSize.Width, d2dRenderTarget.PixelSize.Height));
							// 提交，输出到文件。
							bitmapFrameEncode.Commit();
							encoder.Commit();
						}
					}
				}
			}

			// ===== 清理资源 =====
			d2dRenderTarget.Dispose();

			gaussBlurEffect.Dispose();
			pointSpecularEffect.Dispose();
			composite.Dispose();
			aritheticComposite.Dispose();

			whiteBrush.Dispose();
			redBrush.Dispose();
			textFormat.Dispose();

			dwFactory.Dispose();
			d2dContext.Dispose();
			imagingFactory.Dispose();
			d2dDevice.Dispose();
			dxgiDevice.Dispose();
			d3dDevice.Dispose();
			defaultDevice.Dispose();

			// 显示最终的结果。
			Process.Start(outputPath);
		}
	}
}